import numpy as np
import cv2

from gbvision.exceptions import AbstractMethodCallingException


class StreamReceiver:
    """
    this is an abstract receiver that receives stream from a broadcast receiver
    this class should not be instanced but inherited from
    """

    def __init__(self, shape=(0, 0), fx: float = 1.0, fy: float = 1.0):
        """
        initializes a stream receiver
        :param shape: optional, the shape (x, y) of the sent frame, when set to something other then (0, 0) it overrides
        the fx and fy parameters, when set to (0, 0) it is not used
        :param fx: ratio between width of the read frame to the width of the frame returned, default is 1 (same width)
        :param fy: ratio between height of the read frame to the height of the frame returned,
         default is 1 (same height)
        """
        self.shape = shape
        self.fx = fx
        self.fy = fy

    def get_frame(self) -> np.ndarray:
        """
        reads a frame from the stream and returns in
        :returns: the frame read as a numpy array
        """
        raise AbstractMethodCallingException()

    def _prep_frame(self, frame):
        """
        prepares the frame to be returned and used
        resize and convert to bgr channeled image
        :param frame: the frame to be prepared
        :return: the frame after preparations
        """
        if frame is None:
            return frame
        if len(frame.shape) < 3:
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
        frame = cv2.resize(frame, self.shape, fx=self.fx, fy=self.fy)
        return frame



from gbvision.utils.pipeline import PipeLine
import numpy as np
import cv2


def circle_collision(center1, r1, center2, r2):
    return (center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2 < (r1 + r2) ** 2

@PipeLine
def filter_inner_circles(circles):
    filtered_circles = []
    for i, circle in enumerate(circles):
        circle_invalid = False
        for j in range(i):
            circle_invalid = circle_collision(circle[0], circle[1], circles[j][0], circles[j][1])
            if circle_invalid:
                break
        if not circle_invalid:
            filtered_circles.append(circle)

    return filtered_circles


def rect_collision(r1, r2):
    return not (r1[0] > r2[0] + r2[2] or
                r1[0] + r1[2] < r1[0] or
                r1[1] > r2[1] + r2[3] or
                r1[1] + r1[3] < r2[1])

@PipeLine
def filter_inner_rects(rects):
    filtered_rects = []
    for i, rect in enumerate(rects):
        rect_invalid = False
        for j in range(i):
            rect_invalid = rect_collision(rect, rects[j])
            if rect_invalid:
                break
        if not rect_invalid:
            filtered_rects.append(rect)
    return filtered_rects


def convex_shape_collision(shape1, shape2):
    """
    detects collision between two convex shapes
    Note: if you are uncertain if a shape is convex, use convex_hull on it, it will make it convex
    :param shape1: the first shape, as a contour
    :param shape2: the second shape, as a contour
    :return:
    """
    # shape1_lines = shape1 - np.roll(shape1, 1, axis=0)
    # shape2_lines = shape2 - np.roll(shape2, 1, axis=0)
    # shape1_normals = shape1_lines[:, :, ::-1].reshape(-1, 2, 1) * np.array([[1], [-1]])
    # shape1_normals = shape1_normals.reshape(1, *shape1_normals.shape)
    # shape2_normals = shape2_lines[:, :, ::-1].reshape(-1, 2, 1) * np.array([[1], [-1]])
    # shape2_normals = shape2_normals.reshape(1, *shape2_normals.shape)
    # dot1_1 = shape1.dot(shape1_normals)
    # dot1_2 = shape1.dot(shape2_normals)
    # dot2_1 = shape2.dot(shape1_normals)
    # dot2_2 = shape2.dot(shape2_normals)

    for shape in [shape1, shape2]:
        for idx, edge1 in enumerate(shape):
            edge2 = shape[idx % len(shape1)]

            normal = (edge1 - edge2)[::-1].T
            normal[0] = -normal[0]

            min_1 = max_1 = None
            for edge in shape1:
                projected = edge.dot(normal)[0, 0]
                if min_1 is None or projected < min_1:
                    min_1 = projected

                if max_1 is None or projected > max_1:
                    max_1 = projected

            min_2 = max_2 = None
            for edge in shape2:
                projected = edge.dot(normal)[0, 0]
                if min_2 is None or projected < min_2:
                    min_2 = projected

                if max_2 is None or projected > max_2:
                    max_2 = projected

            if max_1 < min_2 or max_2 < min_1:
                return False

    return True


@PipeLine
def filter_inner_convex_shapes(shapes):
    filtered_shapes = []
    for i, shape in enumerate(shapes):
        shape_invalid = False
        for j in range(i):
            shape_invalid = convex_shape_collision(shape, shapes[j])
            if shape_invalid:
                break
        if not shape_invalid:
            filtered_shapes.append(shape)
    return filtered_shapes


def rotated_rect_collision(rr1, rr2):
    return convex_shape_collision(np.array([cv2.boxPoints(rr1)]), np.array([cv2.boxPoints(rr2)]))

@PipeLine
def filter_inner_rotated_rects(rotated_rects):
    filtered_rotated_rects = []
    for i, rotated_rect in enumerate(rotated_rects):
        rotated_rect_invalid = False
        for j in range(i):
            rotated_rect_invalid = rotated_rect_collision(rotated_rect, rotated_rects[j])
            if rotated_rect_invalid:
                break
        if not rotated_rect_invalid:
            filtered_rotated_rects.append(rotated_rect)
    return filtered_rotated_rects



import cv2

from .threshold_group import ThresholdGroup


def hls_threshold(frame, params):
    """
    thresholds the image according to HLS values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [lmin lmax]
                                                 [smin smax]
    :return: binary threshold image
    """

    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2HLS)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def hsv_threshold(frame, params):
    """
    thresholds the image according to HSV values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [hmin hmax]
                                                 [smin smax]
                                                 [vmin vmax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def rgb_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [rmin rmax]
                                                 [gmin gmax]
                                                 [bmin bmax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def bgr_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [bmin bmax]
                                                 [gmin gmax]
                                                 [rmin rmax]
    :return: binary threshold image
    """
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def luv_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [lmin lmax]
                                                 [umin umax]
                                                 [vmin vmax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2LUV)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def lab_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [lmin lmax]
                                                 [amin amax]
                                                 [bmin bmax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2LAB)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def yuv_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [ymin ymax]
                                                 [umin umax]
                                                 [vmin vmax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2YUV)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def xyz_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 3x2 matrix of [xmin xmax]
                                                 [ymin ymax]
                                                 [zmin zmax]
    :return: binary threshold image
    """
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2XYZ)
    red, green, blue = params
    return cv2.inRange(frame, (red[0], green[0], blue[0]), (red[1], green[1], blue[1]))


def gray_threshold(frame, params):
    """
    thresholds the image according to RGB values
    :param frame: the image
    :param params: the hls values, 1x2 matrix of [min max]
    :return: binary threshold image
    """
    if len(frame.shape) > 2:
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    return cv2.threshold(frame, params[0][0], params[0][1], cv2.THRESH_BINARY)[1]


THRESHOLD_NAME_TABLE = {
    'BGR': bgr_threshold,
    'RGB': rgb_threshold,
    'HLS': hls_threshold,
    'HSV': hsv_threshold,
    'LUV': luv_threshold,
    'LAB': lab_threshold,
    'YUV': yuv_threshold,
    'XYZ': xyz_threshold,
    'GRAY': gray_threshold
}


class Threshold:
    """
    a class that represents a function that maps from an image to a binary image
    where 255 states that the pixel at the original image was in a range represented by the threshold object
    and 0 states the pixel was out of the range
    for example:
    Threshold([[200, 255], [0, 50], [0, 50]], 'RGB')
    the above threshold represents a relatively red pixel. when an image is filtered by it, every pixel that is
    relatively red will be given the value 255, and every pixel that isn't will be given the value of 0
    """

    def __init__(self, pixel_range, thresh_type='HSV'):
        """
        initializes the threshold
        :param pixel_range: the threshold parameters, as a list of integers
        in the shape of 3x2 for a color image and 1x2 for a gray image
        :param thresh_type: a string, the type of color encoding to transform the image before applying the range test
        binary filter
        can be selected from the given list:
        'BGR': default opencv color encoding (no change)
        'RGB': default opencv color encoding in reverse
        'HLS': hue, luminous, saturation
        'HSV': hue, saturation, value
        'LUV': https://en.wikipedia.org/wiki/CIELUV
        'LAB': https://en.wikipedia.org/wiki/CIELAB_color_space
        'YUV': https://en.wikipedia.org/wiki/YUV
        'XYZ': https://en.wikipedia.org/wiki/CIE_1931_color_space
        'GRAY': grayscale images (single channel), image presented doesn't have to be gray, the threshold will convert
         it
        """
        assert thresh_type.upper() in THRESHOLD_NAME_TABLE
        self.params = pixel_range
        self.type = thresh_type.upper()

    def __len__(self):
        """
        the amount of parameters is equal to the amount of channels
        :return: 1 if the threshold is for grayscale images, 3 if it is for color images
        """
        return len(self.params)

    def __getitem__(self, item: int):
        """
        returns the item'th channel's range
        :param item: the index
        :return: the range of the pixel in the item'th channel
        """
        return self.params[item]

    def __setitem__(self, key: int, value: list or tuple):
        """
        sets the key'th channel's range to value
        :param key: the channel
        :param value: an array of 2 integers, the lower and upper bounds of the pixel
        """
        self.params[key] = value

    def __iter__(self):
        """
        :return: an iterator that iterates through the channel ranges
        """
        return iter(self.params)

    def __call__(self, frame):
        """
        activates the threshold filter on the given image
        :param frame: the image to activate the threshold on
        :return: a binary image, the frame after the threshold filter
        """
        return THRESHOLD_NAME_TABLE[self.type](frame, self.params)

    def __repr__(self):
        return '<%s>' % str(self)

    def __str__(self):
        return '%s Threshold(%s)' % (self.type, self.params)

    def __or__(self, other):
        return ThresholdGroup(self, other, binary_mask=cv2.bitwise_or, default_pixel=0)

    def __and__(self, other):
        return ThresholdGroup(self, other, binary_mask=cv2.bitwise_and, default_pixel=0xFF)



from .stream_camera import Camera, StreamCamera


class CameraList(StreamCamera):
    """
    behaves as both a camera and a list of cameras
    camera list holds in it a list of cameras referenced as the field cameras
    and also a single camera to be the current camera used for every operation on the camera list
    as a single camera
    """

    def __init__(self, cameras: list, select_cam: int = None):
        """
        :param cameras: list of the cameras which will be part of the camera list
        you can also add and remove cameras later using the
        :param select_cam: optional, an initial camera to be selected, if not set default camera is the first
        one in the list
        """
        self.cameras = cameras[:]
        if select_cam is None and len(cameras) > 0:
            select_cam = 0
        self.camera: Camera or StreamCamera = self.cameras[select_cam] if select_cam is not None else None

    def __getitem__(self, item: int):
        """
        returns the camera at the index
        :param item: the index
        :return: the camera
        """
        return self.cameras[item]

    def __setitem__(self, item: int, value: Camera):
        """
        sets the camera at the index to the new camera
        :param item: the index
        :param value: the new camera
        """
        self.cameras[item] = value

    def select_camera(self, index: int):
        """
        sets the selected camera to be the camera at the index
        :param index: the new selected camera's index
        """
        self.camera = self.cameras[index]

    def __delitem__(self, item: int):
        """
        deletes the camera at the index
        :param item:
        """
        if self.camera is self.cameras[item]:
            self.camera = None
        del self.cameras[item]

    def __iter__(self):
        """
        :return: an iterator that iterates through all the cameras
        """
        return iter(self.cameras)

    def read(self, image=None, foreach=False):
        if foreach:
            return [cam.read(image=image) for cam in self.cameras]
        return self.camera.read(image=image)

    def is_opened(self, foreach=False):
        if foreach:
            return list(map(lambda x: x.is_opened(), self.cameras))
        return self.camera.is_opened()

    def add_camera(self, cam: Camera):
        """
        adds a new camera to the end of the list
        :param cam: the new camera
        """
        self.cameras.append(cam)

    def release(self, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.release()
        else:
            self.camera.release()
            self.camera = None

    def default(self):
        """
        sets the selected camera to the default camera
        """
        self.camera = self.cameras[0] if len(self.cameras) > 0 else None

    def set_exposure(self, exposure, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.set_exposure(exposure)
        else:
            return self.camera.set_exposure(exposure)

    def toggle_auto_exposure(self, auto, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.toggle_auto_exposure(auto)
        else:
            return self.camera.set_auto_exposure(auto)

    @property
    def focal_length(self):
        return self.camera.focal_length

    @property
    def fov(self):
        return self.camera.fov

    @property
    def data(self):
        return self.camera.data

    def resize(self, x_factor, y_factor, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.resize(x_factor, y_factor)
        else:
            self.camera.resize(x_factor, y_factor)

    def rescale(self, factor, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.rescale(factor)
        else:
            self.camera.rescale(factor)

    def set_frame_size(self, width, height, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.set_frame_size(width, height)
        else:
            self.camera.set_frame_size(width, height)

    def toggle_stream(self, should_stream, foreach=False):
        if foreach:
            for cam in self.cameras:
                if isinstance(cam, StreamCamera):
                    cam.toggle_stream(should_stream)
        else:
            if isinstance(self.camera, StreamCamera):
                self.camera.toggle_stream(should_stream)

    def is_streaming(self, foreach=False):
        if foreach:
            return list(map(lambda x: x.is_streaming(), filter(lambda x: isinstance(x, StreamCamera), self.cameras)))
        return self.camera.is_streaming()

    @property
    def width(self):
        return self.camera.width

    @property
    def height(self):
        return self.camera.height



from .window import Window
from .stream_window import StreamWindow
from .feed_window import FeedWindow
from .camera_window import CameraWindow
from .drawing_tools import DrawCircles, DrawContours, DrawEllipses, DrawRects, DrawRotatedRects
from .recording_feed_window import RecordingFeedWindow



import pickle
import socket
import struct

import cv2

from .stream_receiver import StreamReceiver


class TCPStreamReceiver(StreamReceiver):
    """
    this class uses TCP to receive a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the server and the receiver is the client
    """

    def __init__(self, ip: str, port: int, shape=(0, 0), fx: float = 1.0, fy: float = 1.0):
        """
        initializes the stream receiver
        :param ip: the IPv4 address of the stream broadcaster, for example '10.45.90.8'
        :param port: the port which TCP should use
        """
        StreamReceiver.__init__(self, shape=shape, fx=fx, fy=fy)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_addr = (ip, port)
        self.socket.connect(self.server_addr)
        self.payload_size = struct.calcsize("I")
        self.data = b''

    def get_frame(self):
        while len(self.data) < self.payload_size:
            self.data += self.socket.recv(4096)

        packed_msg_size = self.data[:self.payload_size]

        self.data = self.data[self.payload_size:]

        msg_size = struct.unpack("I", packed_msg_size)[0]

        while len(self.data) < msg_size:
            self.data += self.socket.recv(4096)

        frame_data = self.data[:msg_size]
        self.data = self.data[msg_size:]
        frame = pickle.loads(frame_data)
        if frame is None:
            return None
        frame = cv2.imdecode(frame, -1)
        return self._prep_frame(frame)



import numpy as np
from copy import deepcopy


class CameraData:
    """
    describes constant about a camera in it's default state used to approximate distance
    between the camera and an object seen in a frame
    """

    def __init__(self, focal_length, fov, yaw_angle=0, pitch_angle=0, roll_angle=0, x_offset=0, y_offset=0, z_offset=0):
        """

        :param focal_length: the focal length of the camera at it's default state, in units of pixels
        can be described as the square root of the amount of pixels an object takes on a frame, multiplied by it's
        distance from the camera and divided by the square root of it's surface
        FOCAL_LENGTH = sqrt(P) * D / sqrt(W*H)
        where P is the amount of pixels in the frame representing the object,
        D is the real life distance between the object and the camera
        W is the real life width of the object
        H is the real life height of the object
        note that this is a constant, whatever object you choose to use, this formula will yield the same result
        :param fov:
        half the viewing angle of the camera (field of view) in radians, can be calculated by placing an object in front
        of the camera, so that the entire object is captured and it's center is at the frame's center.
        the tangent of the angle can be described as the width of the object in real life, divided by the
        product of the object's distance from the camera in real life and the ratio between the width of the frame
        in pixels and the width of the object in the frame, also in pixels
        tan(FOV) = (Wm) / (D * (Wp/Wf))
        where Wm is the real life width of the object
        D is the real life distance between the object and the camera
        Wp is the width of the object in the frame (pixels unit)
        Wf is the width of the frame (pixels unit)
        to calculate the FOV just apply the inverse tangent
        FOV = arctan(tan(FOV))
        :param yaw_angle:
        the clockwise yaw angle (in radians) in which the camera is rotated, the yaw angle is the angle around the y axis,
        it's output only affects the x and z axises.
        set this variable when the camera is rotated around the y axis and you want the output of finder functions
        to represent the original space, rather then the rotated one.
        :param pitch_angle:
        the clockwise pitch angle (in radians) in which the camera is rotated, the pitch angle is the angle around the x axis,
        it's output only affects the y and z axises.
        set this variable when the camera is rotated around the x axis and you want the output of finder functions
        to represent the original space, rather then the rotated one.
        :param roll_angle:
        the clockwise roll angle (in radians) in which the camera is rotated, the roll angle is the angle around the z axis,
        it's output only affects the x and y axises.
        set this variable when the camera is rotated around the z axis and you want the output of finder functions
        to represent the original space, rather then the rotated one.
        :param x_offset:
        the x offset in which the camera is placed
        the distance from the measuring point (usually the center of the robot) to the camera on the x axis (left/right),
        if the camera is to the right this should be positive and if it is left this should be negative
        :param y_offset:
        the y offset in which the camera is placed
        the distance from the measuring point to the camera on the y axis (up/down), if the camera is above the measuring point
        this variable should be positive and if it is below this should be negative
        :param z_offset:
        the z offset in which the camera is placed
        the distance from the measuring point to the camera on the z axis (depth), if the camera is placed outer then the measuring point
        this variable should be positive and if it is inner this should be negative
        """
        self.focal_length = focal_length
        self.fov = fov
        sin, cos = np.sin(yaw_angle), np.cos(yaw_angle)
        rotation_matrix = np.array([[cos, 0, sin],
                                    [0, 1, 0],
                                    [-sin, 0, cos]])
        sin, cos = np.sin(pitch_angle), np.cos(pitch_angle)
        rotation_matrix = rotation_matrix.dot(np.array([[1, 0, 0],
                                                        [0, cos, -sin],
                                                        [0, sin, cos]]))
        sin, cos = np.sin(roll_angle), np.cos(roll_angle)
        rotation_matrix = rotation_matrix.dot(np.array([[cos, -sin, 0],
                                                        [sin, cos, 0],
                                                        [0, 0, 1]]))
        self.rotation_matrix = rotation_matrix
        self.offset = np.array([x_offset, y_offset, z_offset])

    def rotate_yaw(self, angle):
        sin, cos = np.sin(angle), np.cos(angle)
        self.rotation_matrix = self.rotation_matrix.dot(np.array([[cos, 0, sin],
                                                                  [0, 1, 0],
                                                                  [-sin, 0, cos]]))
        return self

    def rotate_pitch(self, angle):
        sin, cos = np.sin(angle), np.cos(angle)
        self.rotation_matrix = self.rotation_matrix.dot(np.array([[1, 0, 0],
                                                                  [0, cos, -sin],
                                                                  [0, sin, cos]]))
        return self

    def rotate_roll(self, angle):
        sin, cos = np.sin(angle), np.cos(angle)
        self.rotation_matrix = self.rotation_matrix.dot(np.array([[cos, -sin, 0],
                                                                  [sin, cos, 0],
                                                                  [0, 0, 1]]))
        return self

    def copy(self):
        return deepcopy(self)



import pickle
import socket
import struct

import cv2

from gbvision.constants.net import LOCAL_SERVER_IP
from .stream_receiver import StreamReceiver


class UDPStreamReceiver(StreamReceiver):
    """
    this class uses UDP to receive a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the client and the receiver is the server
    WARNING: do not use this class to send large images, udp has a limited packet size
    """

    def __init__(self, port: int, shape=(0, 0), fx: float = 1.0, fy: float = 1.0):
        """

        :param port: the port which udp should use
        """
        StreamReceiver.__init__(self, shape=shape, fx=fx, fy=fy)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_addr = (LOCAL_SERVER_IP, port)
        self.socket.bind(self.server_addr)
        self.payload_size = struct.calcsize("I")
        self.data = b''

    def get_frame(self):
        self.data += self.socket.recv(2 ** 20)

        packed_msg_size = self.data[:self.payload_size]

        self.data = self.data[self.payload_size:]

        msg_size = struct.unpack("I", packed_msg_size)[0]

        frame_data = self.data[:msg_size]
        self.data = self.data[msg_size:]
        frame = pickle.loads(frame_data)
        if frame is None:
            return None
        frame = cv2.imdecode(frame, -1)
        return self._prep_frame(frame)



from .tcp_stream_receiver import TCPStreamReceiver
from threading import Thread

class AsyncTCPStreamReceiver(TCPStreamReceiver):
    """
    a tcp streamer that reads from the socket on another thread
    recommended for use when multiple threads are reading from the streamer at the same time
    and also when the reading thread gets paused a lot
    for example when running a window that get's paused or moved around a lot
    """
    def __init__(self, ip: str, port: int, shape=(0, 0), fx=1.0, fy=1.0):
        TCPStreamReceiver.__init__(self, ip=ip, port=port, shape=shape, fx=fx, fy=fy)
        self._frame = None
        self._thread = Thread(target=self._receive_thread_function)
        self._thread.start()

    def _receive_thread_function(self):
        while True:
            self._frame = TCPStreamReceiver.get_frame(self)

    def get_frame(self):
        return self._frame